(** Multisets of elements. *)

open Logic

module type PARAM =
  sig
    val toks_sep : Token.t list
    val toks_begin : Token.t list
    val toks_end : Token.t list
    val toks_empty : Token.t list
    val no_begin_end_when_singleton : bool
  end

module Make (Param : PARAM) (E : T) =
  struct
    include Default
    let props =
      fixpoint
	(fun () ->
	  let e = E.props () in
	  {no_props "Multiset" with
	   df = e.df;
	   st' = e.st';
	   po_entails = e.po_entails;
	   cs_entails = e.cs_entails;
	   cp'_entails = reqand [e.cp'_entails; e.sg'];
	   cp_top = isok;
	   cs_bot = isok;
	   cs_conj = isok;
	   cp_conj = isok;
	   cs_disj = isok;
	   cp_disj = isok;
	 })

    type t = (int * E.t) list

    let feat f = List.for_all (fun (n,e) -> E.feat e) f

    let desc f = List.for_all (fun (n,e) -> E.desc e) f

    let isvalue _ = false

    let rec insert_nelt : int * E.t -> t -> t =
      fun (n,e) -> function
	| [] -> [(n,e)]
	| (n',e')::f ->
	    if E.entails e e' & E.entails e' e
	    then (n+n',e')::f
	    else (n',e')::insert_nelt (n,e) f

    let rec size : t -> int =
      fun l ->
	List.fold_left (fun res (n,_) -> res+n) 0 l

    open Param
    let (++) toks p = Syntax.parse_tokens_and toks p
    let (~~) toks = Syntax.parse_tokens toks

    let rec parse = parser
      | [<f = toks_begin ++ parse_suite_elt>] -> List.rev f
      | [<f = parse_singleton>] -> f
      |	[<_ = ~~ toks_empty>] -> []
    and parse_suite_elt = parser
      |	[<ne = parse_nelt; f = parse_suite_sep>] -> insert_nelt ne f
      | [<_ = ~~ toks_end>] -> []
    and parse_suite_sep = parser
      | [<f = toks_sep ++ parse_suite_elt>] -> f
      | [<_ = ~~ toks_end>] -> []
    and parse_singleton str =
      if Param.no_begin_end_when_singleton
      then match str with parser [<ne = parse_nelt>] -> [ne]
      else raise Stream.Failure
    and parse_nelt = parser
      | [<'Token.Nat n; e = E.parse>] -> (n,e)
      | [<e = E.parse>] -> (1,e)

    let rec print = function
      | [] -> toks_empty
      | [ne] when Param.no_begin_end_when_singleton -> print_nelt ne
      | f -> toks_begin @ print_suite f
    and print_suite = function
      |	[] -> toks_end
      | [ne] -> print_nelt ne @ toks_end
      |	ne::f -> print_nelt ne @ toks_sep @ Token.PP_tilda :: print_suite f
    and print_nelt = function
      | (1,e) -> E.print e
      | (n,e) -> Token.Nat n :: Token.PP_tilda :: E.print e


    let top () = []

    let rec entails_loop ress = function
      | [] -> true
      | (0,y)::[] -> true
      | (0,y)::(n',y')::nys ->
	  let ress' = List.map (fun (_,n,x) -> (E.entails x y', n, x)) ress in
	  entails_loop ress' ((n',y')::nys)
      | (n,y)::nys ->
	  (* in ress, the booleans indicates which elements are subsumed by y *)
	  List.exists
	    (fun ress' -> entails_loop ress' ((n-1,y)::nys))
	    (entails_ress_next ress)
    and entails_ress_next = function
      | [] -> []
      | (sub,n,x)::l ->
	  assert (n>0);
	  let res1 = List.map (fun l' -> (sub,n,x)::l') (entails_ress_next l) in
	  if sub then
	    if n=1
	    then l :: res1
	    else ((sub,n-1,x)::l) :: res1
	  else res1

    let rec entails f g =
      match g with
      | [] -> true
      | (n,y)::g' ->
	  let ress = List.map (fun (n,x) -> assert (n>0); (E.entails x y, n, x)) f in
	  entails_loop ress g

    let share f g = true

    let rec features = function
      | [] -> LSet.singleton (true,[])
      |	(n,e)::f ->
	  LSet.add
	    (true,[])
	    (List.fold_right
	       (fun (vis,e') res ->
		 Common.fold_for
		   (fun i res' -> LSet.add (vis,[(i,e')]) res')
		   1 n res)
	       ((*(true,e) ::*) E.features e)
	       (features f))

    let terms f =
      List.fold_left
	(fun res (n,e) -> E.terms e @ res)
	[]
	f

    let axiom f g_opt =
      match f, g_opt with
      | [(n,e)], None -> E.axiom e None
      | [(n,e)], Some [(n',e')] when n>=n' -> E.axiom e (Some e')
      | _, _ -> []

(* conjunction to be proved consistent and complete
    let conj_nelt ld (n,e) =
      let ld' =
	Common.mapfilter
	  (fun (m,d) ->
	    try Some (max m n, E.conj d e)
	    with Not_found ->
	      if n >= m && E.entails e d
	      then None
	      else Some (m,d))
	  ld in
      if entails ld' [(n,e)]
      then ld'
      else (n,e)::ld

    let conj ld le =
      List.fold_left
	(fun res (n,e) -> conj_nelt res (n,e))
	ld
	le
*)


    let add_nelt ld (n,e) =
      let ld' =
	Common.mapfilter
	  (fun (m,d) ->
	    try Some (max m n, E.add d e)
	    with Not_found ->
	      if n >= m && E.entails e d
	      then None
	      else Some (m,d))
	  ld in
      if entails ld' [(n,e)]
      then ld'
      else (n,e)::ld

    let add ld = function
      | [(n,e)] -> add_nelt ld (n,e)
      | _ -> raise Not_found

    let sub_nelt ld (n,e) =
      Common.mapfilter
	(fun (m,d) ->
	  try Some (m, E.sub d e)
	  with Not_found ->
	    if m >= n && E.entails d e
	    then if n-1 > 0 then Some (n-1, d) else None
	    else Some (m,d))
	ld

    let sub ld = function
      | [(n,e)] -> sub_nelt ld (n,e)
      | _ -> raise Not_found

    let list_fold_left_rest f e l =
      let rec aux f e ll lr =
	match lr with
	| [] -> e
	| x::xs -> aux f (f e x (List.rev_append ll xs)) (x::ll) xs in
      aux f e [] l

    (* we assume E.gen is complete at first call
       and we ensure this same property *)
    let rec gen =
      let gen_depth_init = 1 in
      let depth = ref gen_depth_init in
      let egen =
	let tbl : (E.t * E.t, E.t LSet.t) Hashtbl.t = Hashtbl.create 20 in (* tabulation of calls to E.gen *)
	fun ef eg ->
	  try Hashtbl.find tbl (ef, eg)
	  with Not_found ->
	    decr depth;
	    let exs = E.gen ef eg [] in
	    incr depth;
	    Hashtbl.add tbl (ef, eg) exs;
	    exs in
      let rec aux f g = (* f is shorter than g --> we try to map each element of f to an element of g *)
	match f with
	| [] -> LSet.singleton []
	| (nf,ef)::f' ->
	    let xs0 = [] in
	    let xs1 = (* pairing some of (nf,ef) *)
	      list_fold_left_rest
		(fun res (ng,eg) g' ->
		  let exs = egen ef eg in
		  if LSet.is_empty exs
		  then res
		  else
		    Common.fold_for (fun n res0 ->
		      let f' = let nf' = nf - n in if nf' = 0 then f' else (nf',ef)::f' in
		      let g' = let ng' = ng - n in if ng' = 0 then g' else (ng',eg)::g' in
		      let xs' = aux f' g' in
		      List.fold_left (fun res' ex' ->
			List.fold_left (fun res'' x' ->
			  Common.insert_max (fun x y -> entails y x) (insert_nelt (n,ex') x') res''
			  ) res' xs'
			) res0 exs
		      ) 1 (min nf ng) res 
		) xs0 g in
	    let xs2 = (* forgetting (nf,ef) *)
	      let xs' = aux f' g in
	      List.fold_left (fun res x' ->
		Common.insert_max (fun x y -> entails y x) x' res
		) xs1 xs' in
	    LSet.of_list xs2 in
      fun f g hs ->
	if !depth = 0
	then LSet.empty ()
	else
	  let xs =
	    if size f <= size g
	    then aux f g
	    else aux g f in
	  if LSet.is_empty xs
	  then begin incr depth; gen f g hs end
	  else begin depth := gen_depth_init; List.filter (fun x -> not (List.exists (fun h -> entails h x) hs)) xs end

    let rec simpl = function
      | [] -> [<>]
      | (n,e)::l ->
	  [<'l; Common.stream_map (fun e' -> (n,e')::l) (E.simpl e);
	    Common.stream_map (fun l' -> (n,e)::l') (simpl l)>]

  end

